/**
 * Superclass of all errors
 */
export declare class QueryBuilderError extends Error {
}
/**
 * Raised when the builder API is misused.
 *
 * Not all usages are fully covered by the type system,
 * e.g. calling .selectAsJsonAgg multiple times on a table or not having a
 * primary key for a table.
 */
export declare class QueryBuilderUsageError extends QueryBuilderError {
}
/**
 * Thrown upon validation fails when using column validation.
 *
 * Either directly by the builtin colum types or as a wrapper around checking
 * insert data to add column and row information to exceptions thrown by
 * columns runtypes.
 */
export declare class QueryBuilderValidationError extends QueryBuilderError {
    originalError?: Error;
    table?: string;
    column?: string;
    rowNumber?: number;
    row?: any;
    constructor(message?: string, table?: string, column?: string, rowNumber?: number, row?: string, originalError?: Error);
}
/**
 * Thrown by specialized fetch and update functions.
 *
 * E.g. if fetchOne, fetchExactlyOne, updateOne or updateExactlyOne
 * encounter more than 1 row.
 */
export declare class QueryBuilderResultError extends QueryBuilderError {
    constructor(message?: string);
}
