'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var assert = require('assert');
var assert__default = _interopDefault(assert);
var util = require('util');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Date.prototype.toString.call(Reflect.construct(Date, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _construct(Parent, args, Class) {
  if (isNativeReflectConstruct()) {
    _construct = Reflect.construct;
  } else {
    _construct = function _construct(Parent, args, Class) {
      var a = [null];
      a.push.apply(a, args);
      var Constructor = Function.bind.apply(Parent, a);
      var instance = new Constructor();
      if (Class) _setPrototypeOf(instance, Class.prototype);
      return instance;
    };
  }

  return _construct.apply(null, arguments);
}

function _isNativeFunction(fn) {
  return Function.toString.call(fn).indexOf("[native code]") !== -1;
}

function _wrapNativeSuper(Class) {
  var _cache = typeof Map === "function" ? new Map() : undefined;

  _wrapNativeSuper = function _wrapNativeSuper(Class) {
    if (Class === null || !_isNativeFunction(Class)) return Class;

    if (typeof Class !== "function") {
      throw new TypeError("Super expression must either be null or a function");
    }

    if (typeof _cache !== "undefined") {
      if (_cache.has(Class)) return _cache.get(Class);

      _cache.set(Class, Wrapper);
    }

    function Wrapper() {
      return _construct(Class, arguments, _getPrototypeOf(this).constructor);
    }

    Wrapper.prototype = Object.create(Class.prototype, {
      constructor: {
        value: Wrapper,
        enumerable: false,
        writable: true,
        configurable: true
      }
    });
    return _setPrototypeOf(Wrapper, Class);
  };

  return _wrapNativeSuper(Class);
}

/**
 * Superclass of all errors
 */
var QueryBuilderError =
/*#__PURE__*/
function (_Error) {
  _inheritsLoose(QueryBuilderError, _Error);

  function QueryBuilderError() {
    return _Error.apply(this, arguments) || this;
  }

  return QueryBuilderError;
}(
/*#__PURE__*/
_wrapNativeSuper(Error));
/**
 * Raised when the builder API is misused.
 *
 * Not all usages are fully covered by the type system,
 * e.g. calling .selectAsJsonAgg multiple times on a table or not having a
 * primary key for a table.
 */

var QueryBuilderUsageError =
/*#__PURE__*/
function (_QueryBuilderError) {
  _inheritsLoose(QueryBuilderUsageError, _QueryBuilderError);

  function QueryBuilderUsageError() {
    return _QueryBuilderError.apply(this, arguments) || this;
  }

  return QueryBuilderUsageError;
}(QueryBuilderError);
/**
 * Thrown upon validation fails when using column validation.
 *
 * Either directly by the builtin colum types or as a wrapper around checking
 * insert data to add column and row information to exceptions thrown by
 * columns runtypes.
 */

var QueryBuilderValidationError =
/*#__PURE__*/
function (_QueryBuilderError2) {
  _inheritsLoose(QueryBuilderValidationError, _QueryBuilderError2);

  function QueryBuilderValidationError(message, table, column, rowNumber, row, originalError) {
    var _this;

    _this = _QueryBuilderError2.call(this, message) || this;
    _this.name = 'QueryBuilderValidationError';
    _this.table = table;
    _this.column = column;
    _this.rowNumber = rowNumber;
    _this.row = row;
    _this.originalError = originalError;
    return _this;
  }

  return QueryBuilderValidationError;
}(QueryBuilderError);
/**
 * Thrown by specialized fetch and update functions.
 *
 * E.g. if fetchOne, fetchExactlyOne, updateOne or updateExactlyOne
 * encounter more than 1 row.
 */

var QueryBuilderResultError =
/*#__PURE__*/
function (_QueryBuilderError3) {
  _inheritsLoose(QueryBuilderResultError, _QueryBuilderError3);

  function QueryBuilderResultError(message) {
    var _this2;

    _this2 = _QueryBuilderError3.call(this, message) || this;
    _this2.name = 'QueryBuilderResultError';
    return _this2;
  }

  return QueryBuilderResultError;
}(QueryBuilderError);

function identity(value) {
  return value;
}

function column(sqlName, validator, // checks and or converts the datatype when inserting a column
fromJson) {
  if (!validator) {
    return new Column({
      name: sqlName,
      columnValue: identity,
      fromJson: fromJson
    });
  }

  return new Column({
    name: sqlName,
    columnValue: validator,
    fromJson: fromJson
  });
}
/**
 * A column of a table
 */

var Column =
/*#__PURE__*/
function () {
  function Column(params) {
    this.name = params.name;
    this.columnValue = params.columnValue;
    this.fromJson = params.fromJson;
    this.isPrimaryKey = params.isPrimaryKey;
    this.isNullable = params.isNullable;
  } // private but part of ColumnImplementation


  var _proto = Column.prototype;

  _proto.copy = function copy(params) {
    return new Column({
      name: params.name,
      columnValue: this.columnValue,
      fromJson: this.fromJson,
      isPrimaryKey: this.isPrimaryKey,
      isNullable: this.isNullable
    });
  } /// column sql attributes

  /**
   * Mark this column as being the sole or part of the tables primary key.
   *
   * Knowing that a column is a primary key is required to generate correct
   * group-by clauses for json_agg (selectAsJsonAgg) projections.
   */
  ;

  _proto.primary = function primary() {
    this.isPrimaryKey = true;
    return this;
  }
  /**
   * Mark this column has having a default value.
   *
   * Columns with defaults can be ommitted in insert queries.
   * `nullable` values always have null as the default.
   */
  ;

  _proto["default"] = function _default() {
    // cast to any bc we need to change this columns type
    var anyThis = this;
    var columnValue = this.columnValue;

    anyThis.columnValue = function (value) {
      if (value === undefined) {
        // insert filters any undefined columns but the validations runs
        // before that step and is the only part of the column that knowns
        // about it being optional
        return undefined;
      }

      return columnValue(value);
    };

    return anyThis;
  }
  /**
   * Make this column nullable.
   *
   * That means it can hold `null` and also uses null as its default.
   */
  ;

  _proto["null"] = function _null() {
    // cast to any bc we need to change this columns type
    var anyThis = this;
    var fromJson = this.fromJson;
    var columnValue = this.columnValue;

    anyThis.columnValue = function (value) {
      // check for both null and undefined as the default value for nullable
      // columns is always null and undefined in inserts means `use the
      // default`
      if (value === null || value === undefined) {
        return null;
      }

      return columnValue(value);
    }; // explicit tag that this may be null, required to generate `IS NULL`
    // where expressions


    anyThis.isNullable = true;
    anyThis.fromJson = fromJson ? function (value) {
      if (value === null) {
        return null;
      }

      return fromJson(value);
    } : undefined;
    return anyThis;
  } /// built-in column types
  ;

  _proto.checkThatColumnValueIsIdentity = function checkThatColumnValueIsIdentity() {
    if (this.columnValue !== identity) {
      throw new QueryBuilderUsageError("column " + this.name + " - type methods such as .integer(), .string(), and .date() must be called directly after creating a column");
    }
  }
  /**
   * Map this column to a typescript number that must be an integer.
   *
   * Column values being integers is only checked when inserting or updating
   * data.
   */
  ;

  _proto.integer = function integer() {
    var _this = this;

    this.checkThatColumnValueIsIdentity();
    var anyThis = this;

    anyThis.columnValue = function (value) {
      if (typeof value !== 'number') {
        throw new QueryBuilderValidationError("column " + _this.name + " - expected an integer but got: " + util.inspect(value).slice(0, 128));
      }

      if (!Number.isInteger(value) || !Number.isSafeInteger(value)) {
        throw new QueryBuilderValidationError("column " + _this.name + " - expected an integer but got a number with fractions or a non safe integer: " + util.inspect(value));
      }

      return value;
    };

    return anyThis;
  }
  /**
   * Map this column to a string.
   *
   * postgres types: TEXT, VARCHAR
   */
  ;

  _proto.string = function string() {
    var _this2 = this;

    this.checkThatColumnValueIsIdentity();
    var anyThis = this;

    anyThis.columnValue = function (value) {
      if (typeof value !== 'string') {
        throw new QueryBuilderValidationError("column " + _this2.name + " - expected a string but got: " + util.inspect(value).slice(0, 128));
      }

      return value;
    };

    return anyThis;
  }
  /**
   * Map this column to a boolean.
   *
   * postgres type: BOOLEAN
   */
  ;

  _proto["boolean"] = function boolean() {
    var _this3 = this;

    this.checkThatColumnValueIsIdentity();
    var anyThis = this;

    anyThis.columnValue = function (value) {
      if (typeof value !== 'boolean') {
        throw new QueryBuilderValidationError("column " + _this3.name + " - expected a boolean but got: " + util.inspect(value).slice(0, 128));
      }

      return value;
    };

    return anyThis;
  }
  /**
   * Map this column to a date.
   *
   * postgres types: TIMESTAMPTZ
   */
  ;

  _proto.date = function date() {
    var _this4 = this;

    this.checkThatColumnValueIsIdentity();
    var anyThis = this;

    anyThis.columnValue = function (value) {
      if (!(value instanceof Date)) {
        throw new QueryBuilderValidationError("column {this.name} - expected a Date but got: " + util.inspect(value).slice(0, 128));
      }

      return value;
    };

    anyThis.fromJson = function (value) {
      if (value instanceof Date) {
        return value;
      } // postgres serializes timestamps into strings when selected via json functions


      if (typeof value === 'string') {
        return new Date(value);
      }

      throw new QueryBuilderValidationError("column " + _this4.name + " - cannot read Date from " + util.inspect(value).slice(0, 128));
    };

    return anyThis;
  }
  /**
   * Map this column to an json object.
   *
   * Validator should be function that validates the type of the incoming
   * data. Validator is called before inserting or updating and stringifies
   * any data before passing it to postgres.
   *
   * postgres types: JSON, JSONB
   */
  ;

  _proto.json = function json(validator) {
    this.checkThatColumnValueIsIdentity();
    var anyThis = this;

    anyThis.columnValue = function (value) {
      // stringify before insert / update because node-pg does not know when
      // we're inserting / updating json data so we have to pass it as a string
      // (node-pg only stringifies js objects but not arrays or strings)
      // see https://github.com/brianc/node-postgres/issues/442
      return JSON.stringify(validator(value));
    };

    return anyThis;
  };

  _proto.stringUnion = function stringUnion() {
    var _this5 = this;

    for (var _len = arguments.length, elements = new Array(_len), _key = 0; _key < _len; _key++) {
      elements[_key] = arguments[_key];
    }

    this.checkThatColumnValueIsIdentity();
    var index = new Set(elements);
    var anyThis = this;

    anyThis.columnValue = function (value) {
      if (typeof value !== 'string' || !index.has(value)) {
        throw new QueryBuilderValidationError("column " + _this5.name + " - expected a string of " + elements + " but got: " + util.inspect(value).slice(0, 128));
      }

      return value;
    };

    return anyThis;
  };

  _proto["enum"] = function _enum(enumObject) {
    var _this6 = this;

    this.checkThatColumnValueIsIdentity();
    var valueIndex = new Set(Object.entries(enumObject).filter(function (_ref) {
      var k = _ref[0];

      if (/^\d+$/.test(k) && !isNaN(Number(k))) {
        // key is a number so its value is part of typescripts automatic reverse mapping for number enums
        // makes sense because numbers can never be valid enum keys anyway
        // https://www.typescriptlang.org/docs/handbook/enums.html#reverse-mappings
        return false;
      }

      return true;
    }).map(function (_ref2) {
      var v = _ref2[1];
      return v;
    }));
    var anyThis = this;

    anyThis.columnValue = function (value) {
      // reverse lookup for number enums
      if (typeof value === 'number' && enumObject[value] !== undefined) {
        return value;
      }

      if (typeof value !== 'string' || !valueIndex.has(value)) {
        throw new QueryBuilderValidationError("column " + _this6.name + " - expected a member of the enum " + util.inspect(enumObject, {
          compact: true,
          breakLength: Infinity
        }) + " but got: " + util.inspect(value).slice(0, 128));
      }

      return value;
    };

    return anyThis;
  };

  return Column;
}();
/**
 * Return the implementation of a column.
 *
 * The implementation is used to create queries at runtime.
 * The column type is used provide type checking at compile type.
 */

function getColumnImplementation(column) {
  // thanks to ts private fields being just syntax sugar, we can write this
  // function as a simple identity with a type cast
  return column;
}

var tableImplementationSymbol =
/*#__PURE__*/
Symbol('tableImplementation');
/**
 * Base object for all tables implementing the TableProjectionMethods
 *
 * The implementation is hidden anyway thus we're free to use lots of any's.
 */

var TableImplementation =
/*#__PURE__*/
function () {
  function TableImplementation(tableName, tableColumns) {
    this.jsonAggCanaryColumnName = '__typesafe_query_builder_canary_column';
    this.tableName = tableName;
    this.tableColumns = tableColumns;
    this[tableImplementationSymbol] = true;
  }

  var _proto = TableImplementation.prototype;

  _proto.debugInfo = function debugInfo() {
    var isSubQuery = this.tableQuery ? ' (subquery)' : '';
    return "" + this.tableName + isSubQuery;
  };

  _proto.copy = function copy() {
    var res = new TableImplementation(this.tableName, this.tableColumns);
    res.selected = this.selected;
    res.renamed = this.renamed;
    res.projection = this.projection;
    res.referencedColumn = this.referencedColumn;
    res.tableQuery = this.tableQuery;
    return res;
  };

  _proto.createTableColumn = function createTableColumn(name) {
    var res = this.copy();

    if (this.tableColumns[name] === undefined) {
      assert__default.fail("column " + name + " does not exist on table " + this.tableName);
    }

    res.referencedColumn = name;
    return res;
  } // serving the actual Table<T,S>
  ;

  _proto.getTableProxy = function getTableProxy() {
    var _this = this;

    return new Proxy(this, {
      get: function get(_target, prop, _receiver) {
        // TableProjectionMethods
        if (prop === 'select' || prop === 'selectAs' || prop === 'selectAsJson' || prop === 'selectAsJsonAgg' || prop === 'selectWithout' || prop === 'column') {
          return _this[prop];
        } // TableColumn


        if (typeof prop === 'string' && _this.tableColumns[prop]) {
          return _this.createTableColumn(prop);
        } // Access to this implementation (used to build the sql query)


        if (prop === tableImplementationSymbol) {
          return _this;
        }

        return undefined;
      }
    });
  } // return the columns that are selected / renamed/ json-projected by this
  // table as Column objects to be used in new tables that cover a subselect
  ;

  _proto.getColumns = function getColumns() {
    var _this2 = this;

    if (this.projection === undefined) {
      var res = {};
      var cols = this.selected || Object.keys(this.tableColumns);
      cols.forEach(function (key) {
        var _this2$renamed;

        var colKey = ((_this2$renamed = _this2.renamed) === null || _this2$renamed === void 0 ? void 0 : _this2$renamed.hasOwnProperty(key)) ? _this2.renamed[key] : key;

        var colImpl = _this2.tableColumns[key].copy({
          name: colKey
        });

        res[colKey] = colImpl;
      });
      return res;
    } else if (this.projection.type === 'jsonBuildObject') {
      var _ref;

      // a single column that validates & fromJson-izes correctly
      var fromJsonConverter = this.getFromJsonRowConverter();
      return _ref = {}, _ref[this.projection.name] = new Column({
        name: this.projection.name,
        columnValue: function columnValue() {
          throw new QueryBuilderUsageError('cannot use subselected tables in inserts / updates');
        },
        fromJson: function fromJson(v) {
          // use the converter in the cols fromJson so that we are free to
          // combine the from json in any way
          fromJsonConverter(v);
          return v;
        }
      }), _ref;
    } else if (this.projection.type === 'jsonAgg') {
      var _ref2;

      // a single column that validates & fromJson-izes the resulting json-agg array correctly
      var _fromJsonConverter = this.getFromJsonRowConverter();

      return _ref2 = {}, _ref2[this.projection.name] = new Column({
        name: this.projection.name,
        columnValue: function columnValue() {
          throw new QueryBuilderUsageError('cannot use subselected tables in inserts / updates');
        },
        fromJson: function fromJson(v) {
          for (var i = 0; i < v.length; i++) {
            _fromJsonConverter(v[i]);
          }

          return v;
        }
      }), _ref2;
    } else {
      assert__default.fail("invalid projection in table " + this.tableName + " - " + this.projection);
    }
  } // return the javascript name of the column, which may have possibly renamed
  // through `selectAs`
  ;

  _proto.getColumnResultKey = function getColumnResultKey(columnKey) {
    if (!this.renamed) {
      return columnKey;
    }

    var key = this.renamed[columnKey];

    if (key === undefined) {
      return columnKey;
    }

    return key;
  } // Return the select expression to fetch this tables selected / renamed /
  // json-projected columns.
  // alias is undefined when generating insert or update statements, otherwise its prepended in front of any accessed table column.
  // Set left join to true to make the returned sql check for null primary
  // keys in order to detect empty left joins and build json objects
  // accordingly.
  ;

  _proto.getSelectSql = function getSelectSql(alias, isLeftJoin) {
    var _this3 = this;

    var selected = this.selected || Object.keys(this.tableColumns);
    var aliasPrefix = alias ? alias + '.' : '';

    if (this.projection === undefined) {
      // default projection (none)
      return selected.map(function (columnKey) {
        var column = _this3.tableColumns[columnKey];

        var columnResultKey = _this3.getColumnResultKey(columnKey);

        return aliasPrefix + "\"" + column.name + "\" AS \"" + columnResultKey + "\"";
      }).join(',');
    } // build a json object with sql


    var jsonBuildObject = this.getJsonBuildObjectExpression(alias); // identify a null row caused by a left join
    // need this as otherwise we would create json objects full with null
    // values e.g.: `{id: null, userName: null}` instead of just `null`

    var jsonIsNotNull = this.getLeftJoinIsNullCanaryColumnSql(alias);

    if (this.projection.type === 'jsonBuildObject') {
      // this is just an optimization to only generate the CASE expresssion when its needed
      if (isLeftJoin) {
        return "(CASE WHEN " + jsonIsNotNull + " THEN " + jsonBuildObject + " ELSE null END) AS \"" + this.projection.name + "\"";
      } else {
        return jsonBuildObject + " AS \"" + this.projection.name + "\"";
      }
    } else if (this.projection.type === 'jsonAgg') {
      // json_agg supports order by
      var jsonAggOrderBy = this.projection.orderBy ? " ORDER BY " + aliasPrefix + "\"" + this.tableColumns[this.projection.orderBy].name + "\" " + (this.projection.direction || '') : ''; // For left joins with missing values, make postgres return an
      // empty json array [] instead of [null]
      // see https://stackoverflow.com/questions/24155190/postgresql-left-join-json-agg-ignore-remove-null
      // To check that the result will be empty, use the tables primary key which never should be null unless left-joined.

      return 'COALESCE(JSON_AGG(' + jsonBuildObject + jsonAggOrderBy + ') FILTER (WHERE ' + jsonIsNotNull + "), '[]') AS \"" + this.projection.name + '"';
    } else {
      assert__default.fail("invalid projection in table " + this.tableName + " - " + this.projection);
    }
  };

  _proto.getTableSql = function getTableSql(alias, ctx, params) {
    var tableSql;

    if (this.tableQuery) {
      // subquery created with Query.table()
      if (this.isJsonAggProjection() || this.isJsonProjection()) {
        // In case that we are a table projection, we need to add a canary
        // column that serves us to check whether a left joined culumn was all
        // nulls so its not included in the json agg array at all.
        // (Cant use `<table> IS NULL` because that would also filter out
        // legitimate `{foo: null}` values from the json agg array.
        // In normal tables we can simply rely on the tables primary key but
        // in subqueries the PK might have been omitted from the result.
        // that column can be static bc. there is only ever a single json
        // agg per table / query anyway.
        tableSql = "(" + this.tableQuery(ctx, params, this.jsonAggCanaryColumnName) + ")";
      } else {
        tableSql = "(" + this.tableQuery(ctx, params) + ")";
      }
    } else {
      // plain table select, use the primary keys to detect nulls
      tableSql = "\"" + this.tableName + "\"";
    }

    return tableSql + " " + alias;
  };

  _proto.getJsonBuildObjectExpression = function getJsonBuildObjectExpression(alias) {
    var _this4 = this;

    var aliasPrefix = alias ? alias + '.' : '';
    return 'JSON_BUILD_OBJECT(' + (this.selected || Object.keys(this.tableColumns)).map(function (columnKey) {
      var column = _this4.tableColumns[columnKey];

      var columnResultKey = _this4.getColumnResultKey(columnKey);

      return "'" + columnResultKey + "', " + aliasPrefix + "\"" + column.name + "\"";
    }).join(',') + ')';
  };

  _proto.getReferencedColumn = function getReferencedColumn() {
    if (!this.referencedColumn) {
      assert__default.fail("'referencedColumn' in table " + this.tableName + " is undefined");
    }

    return this.tableColumns[this.referencedColumn];
  };

  _proto.getReferencedColumnSql = function getReferencedColumnSql(alias) {
    var col = this.getReferencedColumn();
    return alias + ".\"" + col.name + "\"";
  } // return a list of primary column expressions to build "group by" clauses
  ;

  _proto.getPrimaryColumnsSql = function getPrimaryColumnsSql(alias) {
    var aliasPrefix = alias ? alias + '.' : '';
    var res = Object.values(this.tableColumns).filter(function (c) {
      return c.isPrimaryKey;
    }).map(function (pk) {
      return aliasPrefix + "\"" + pk.name + "\"";
    });

    if (!res.length) {
      assert__default.fail("table has no primary columns: " + this.debugInfo());
    }

    return res;
  } // return an sql expression used to test whether this left joined table is null or present
  ;

  _proto.getLeftJoinIsNullCanaryColumnSql = function getLeftJoinIsNullCanaryColumnSql(alias) {
    if (this.tableQuery) {
      // subquery created with Query.table()
      // tableQuery for an explanation of jsonAggCanaryColumnName
      return alias + "." + this.jsonAggCanaryColumnName;
    } else {
      // plain table select, use the primary keys to detect nulls
      var pks = this.getPrimaryColumnsSql(alias);

      if (!pks.length) {
        assert__default.fail("table has no primary columns: " + this.debugInfo());
      }

      return pks.map(function (pkSql) {
        return pkSql + " IS NOT NULL";
      }).join(' AND ');
    }
  } // using the json_agg (via selectAsJsonAgg) function implies a
  // "group by <primary-key-columns>"
  ;

  _proto.isJsonAggProjection = function isJsonAggProjection() {
    return !!(this.projection && this.projection.type === 'jsonAgg');
  };

  _proto.isJsonProjection = function isJsonProjection() {
    return !!(this.projection && this.projection.type === 'jsonBuildObject');
  } // Return a function that converts a flat row of column values according to
  // each selected / renamed columns `fromJson` method.
  // Used internally to build up subselect and projection aware result converters.
  ;

  _proto.getFromJsonRowConverter = function getFromJsonRowConverter() {
    // collect the `fromJson` method from each selected column
    var keys = [];
    var fromJsons = [];
    var allSelectedCols = this.selected || Object.keys(this.tableColumns);

    for (var i = 0; i < allSelectedCols.length; i++) {
      var colKey = allSelectedCols[i];
      var colImpl = this.tableColumns[colKey];

      if (colImpl.fromJson) {
        keys.push(this.getColumnResultKey(colKey)); // possibly renamed colKey

        fromJsons.push(colImpl.fromJson);
      }
    } // not a single column needs a result conversion


    if (fromJsons.length === 0) {
      return function () {};
    }

    return function (row) {
      for (var _i = 0; _i < fromJsons.length; _i++) {
        var key = keys[_i];
        var fromJson = fromJsons[_i];

        if (row[key] === null) {
          // possible null caused by a left join
          // NOTE: actually, this check is only required if we're actually
          // using this table in a left-join, but we don't know it here
          return;
        }

        row[key] = fromJson(row[key]);
      }
    };
  } // turn any Date columns selected via pg json functions into into real dates
  // (because while being selected as json, postgres converts them to strings
  // first so they fit into the json column)
  // so basically this is a custom datatype serialization on top of what
  // node-postgres already provides
  ;

  _proto.getResultConverter = function getResultConverter() {
    // return a function that converts fields of a single row *in-place* (to
    // save on memory allocations)
    var rowConverter = this.getFromJsonRowConverter();

    if (this.projection === undefined) {
      return rowConverter;
    } else if (this.projection.type === 'jsonBuildObject') {
      // plain json projection: all selected columns are grouped below `key`
      var key = this.projection.name;
      return function (row) {
        if (row[key] === null) {
          // null caused by a left join and missing data
          return;
        }

        rowConverter(row[key]);
      };
    } else if (this.projection.type === 'jsonAgg') {
      // json agg -> key is an array of the joined rows
      var _key = this.projection.name;
      return function (row) {
        // row[key] is never null because with json-agg we use coalesce
        // and json-agg where filter to create an empty array instead and
        // remove all null values
        var jsonAggValue = row[_key];

        for (var i = 0; i < jsonAggValue.length; i++) {
          rowConverter(jsonAggValue[i]);
        }
      };
    } else {
      assert__default.fail("invalid projection in table " + this.tableName + " - " + this.projection);
    }
  } // return the names of all columns as they appear in the query result
  // use this to check shadowed columns in query.ts
  ;

  _proto.getResultingColumnNames = function getResultingColumnNames() {
    var _this5 = this;

    if (this.projection) {
      return [this.projection.name];
    }

    var keys = this.selected || Object.keys(this.tableColumns);
    return keys.map(function (k) {
      return _this5.getColumnResultKey(k);
    });
  } /// TableProjectionMethods implementation
  // choose columns to appear in the result.
  ;

  _proto.select = function select() {
    var res = getTableImplementation(this).copy();

    if (res.selected) {
      throw new QueryBuilderUsageError("only a single select call is allowed for a Table (tableName: " + this.tableName);
    }

    for (var _len = arguments.length, keys = new Array(_len), _key2 = 0; _key2 < _len; _key2++) {
      keys[_key2] = arguments[_key2];
    }

    res.selected = keys;
    return res.getTableProxy();
  } // choose columns to *hide* from the result.
  ;

  _proto.selectWithout = function selectWithout() {
    for (var _len2 = arguments.length, keys = new Array(_len2), _key3 = 0; _key3 < _len2; _key3++) {
      keys[_key3] = arguments[_key3];
    }

    var res = getTableImplementation(this).copy();
    res.selected = (res.selected || Object.keys(res.tableColumns)).filter(function (k) {
      return !keys.includes(k);
    });
    return res.getTableProxy();
  } // rename some columns in the result
  ;

  _proto.selectAs = function selectAs(mapping) {
    var res = getTableImplementation(this).copy();

    if (res.renamed) {
      throw new QueryBuilderUsageError("only a single selectAs call is allowed for a  Table (tableName: " + this.tableName);
    }

    res.renamed = mapping;
    return res.getTableProxy();
  } // project all columns into a json object
  ;

  _proto.selectAsJson = function selectAsJson(key) {
    var res = getTableImplementation(this).copy();

    if (res.projection) {
      throw new QueryBuilderUsageError("only a single selectAsJson or selectAsJsonAgg call is allowed for a  Table (tableName: " + this.tableName);
    }

    res.projection = {
      type: 'jsonBuildObject',
      name: key
    };
    return res.getTableProxy();
  } // json_agg projection of a whole table.
  ;

  _proto.selectAsJsonAgg = function selectAsJsonAgg(key, orderBy, direction) {
    var res = getTableImplementation(this).copy();

    if (res.projection) {
      throw new QueryBuilderUsageError("only a single selectAsJson or selectAsJsonAgg call is allowed for a  Table (tableName: " + this.tableName);
    }

    res.projection = {
      type: 'jsonAgg',
      name: key,
      orderBy: orderBy,
      direction: direction
    };
    return res.getTableProxy();
  } // column accessor, in case a column name clashes with a table method
  ;

  _proto.column = function column(name) {
    return this.createTableColumn(name);
  };

  return TableImplementation;
}();
/**
 * Define a relation consisting of typed columns.
 */

function table(tableName, columns) {
  // remove type info from columns to access their private attributes
  var columnImplementations = {};
  Object.keys(columns).forEach(function (k) {
    columnImplementations[k] = getColumnImplementation(columns[k]);
  }); // each table needs at least 1 primary key column

  var hasPrimaryKey = Object.values(columnImplementations).some(function (c) {
    return c.isPrimaryKey;
  });

  if (!hasPrimaryKey) {
    throw new QueryBuilderUsageError("table " + tableName + " does not have any primary keys");
  }

  return new TableImplementation(tableName, columnImplementations).getTableProxy();
}
function getTableImplementation(table) {
  var implementation = table[tableImplementationSymbol];

  if (implementation === undefined) {
    assert__default.fail('table implementation not found');
  }

  if (implementation === true) {
    return table;
  }

  return implementation;
}

function assertNever(x) {
  assert.fail('Unexpected value. Should have been never.');
}

var AliasGenerator =
/*#__PURE__*/
function () {
  function AliasGenerator() {
    this.counter = 0;
    this.aliasMap = {}; // tableName -> alias
  }

  var _proto = AliasGenerator.prototype;

  _proto.getAlias = function getAlias(tableName) {
    if (!this.aliasMap[tableName]) {
      this.aliasMap[tableName] = "a" + this.counter;
      this.counter += 1;
    }

    return this.aliasMap[tableName];
  };

  return AliasGenerator;
}();

var SqlQuery =
/*#__PURE__*/
function () {
  function SqlQuery(ctx) {
    this.ctx = ctx;
    this.from = [];
    this.select = [];
    this.joins = [];
    this.where = [];
    this.groupBy = [];
    this.orderBy = [];
    this.aliasGenerator = new AliasGenerator();
    this.ctx = ctx;
  }

  var _proto2 = SqlQuery.prototype;

  _proto2.getAlias = function getAlias(tableName) {
    return this.aliasGenerator.getAlias(tableName);
  };

  _proto2.addFrom = function addFrom(tableSql) {
    this.from.push(tableSql);
  };

  _proto2.addSelect = function addSelect(selectSql) {
    this.select.push(selectSql);
  };

  _proto2.addJoin = function addJoin(joinType, tableSql2, columnSql1, columnSql2) {
    this.joins.push({
      joinType: joinType,
      tableSql2: tableSql2,
      columnSql1: columnSql1,
      columnSql2: columnSql2
    });
  };

  _proto2.addGroupBy = function addGroupBy(columnSql) {
    var _this$groupBy;

    (_this$groupBy = this.groupBy).push.apply(_this$groupBy, columnSql);
  };

  _proto2.addOrderBy = function addOrderBy(columnSql, direction, nulls) {
    this.orderBy.push([columnSql, direction || '', nulls === 'nullsFirst' ? 'NULLS FIRST' : nulls === 'nullsLast' ? 'NULLS LAST' : ''].join(' '));
  };

  _proto2.setLimit = function setLimit(limit) {
    if (this.limit !== undefined) {
      throw new QueryBuilderUsageError('limit is already present');
    }

    if (limit < 0 || !Number.isInteger(limit)) {
      throw new QueryBuilderUsageError('limit must be > 0 and an integer');
    }

    this.limit = limit;
  };

  _proto2.setOffset = function setOffset(offset) {
    if (this.offset !== undefined) {
      throw new QueryBuilderUsageError('offset is already present');
    }

    if (offset < 0 || !Number.isInteger(offset)) {
      throw new QueryBuilderUsageError('offset must be > 0 and an integer');
    }

    this.offset = offset;
  };

  _proto2.addWhereEq = function addWhereEq(columnSql, paramKey, nullable) {
    var c = columnSql;

    if (nullable) {
      var p1 = this.ctx.getNextParameter(paramKey);
      var p2 = this.ctx.getNextParameter(paramKey); // Null and equals check using a single javascript value.
      // Using a single parameter for null and equals check results in an
      // 'could not determine data type of parameter $N'-error.
      // Still with two parameters a type cast is required for the null-check
      // param so use text as that cast works for most types.

      this.where.push("CASE WHEN " + p1 + "::text IS NULL THEN " + c + " IS NULL ELSE " + c + " = " + p2 + " END");
    } else {
      var p = this.ctx.getNextParameter(paramKey);
      this.where.push(c + " = " + p);
    }
  };

  _proto2.addWhereIn = function addWhereIn(columnSql, paramKey) {
    var c = columnSql;
    var p = this.ctx.getNextParameter(paramKey); // Use an array comparison function (instead of IN) so we can pass
    // parameters as a simple json array in a simple arg without having to
    // know how long the arguments are.
    // see https://www.postgresql.org/docs/current/functions-comparisons.html

    this.where.push(c + " = ANY(" + p + ")");
  } // given the template-string-fragments and some strings, interpolate them
  // into a single resulting string
  ;

  _proto2.createSqlFragmentString = function createSqlFragmentString(literals, params) {
    if (literals.length - 1 !== params.length) {
      assert.fail("SqlQuery: not enough parameters (" + params.length + ") for template literal strings array: " + JSON.stringify(literals));
    }

    var res = [];

    for (var i in params) {
      res.push(literals[i]);
      res.push(params[i]);
    }

    res.push(literals[literals.length - 1]);
    return res.join('');
  };

  _proto2.addWhereSql = function addWhereSql(fragments, columnsSql) {
    var _this = this;

    if (!fragments.length) {
      return;
    }

    if (fragments.length !== columnsSql.length) {
      assert.fail('SqlQuery: fragments and columnsSql must be the same length');
    }

    var sqlExpressions = fragments.map(function (f, i) {
      if (f.column) {
        // fragment contains a column expression ...
        var colSql = columnsSql[i];

        if (!colSql) {
          assert.fail("SqlQuery: columnSql at " + i + " is undefined");
        }

        if (f.paramKey) {
          // ... and a parameter expression ...
          var p = _this.ctx.getNextParameter(f.paramKey);

          if (f.columnFirst) {
            // ... and the col expression comes first
            return _this.createSqlFragmentString(f.literals, [colSql, p]);
          } else {
            // ... and the parameter expression comes first
            return _this.createSqlFragmentString(f.literals, [p, colSql]);
          }
        } else {
          // ... and nothing else
          return _this.createSqlFragmentString(f.literals, [colSql]);
        }
      } else if (f.paramKey) {
        // fragment only contains a parameter expression
        var _p = _this.ctx.getNextParameter(f.paramKey);

        return _this.createSqlFragmentString(f.literals, [_p]);
      } else {
        // constant string only
        return _this.createSqlFragmentString(f.literals, []);
      }
    });
    this.where.push("(" + sqlExpressions.join(' ') + ")");
  };

  _proto2.setLock = function setLock(lockMode) {
    if (lockMode !== 'none' && lockMode !== 'share' && lockMode !== 'update') {
      throw new QueryBuilderUsageError("invalid lock mode parameter: " + lockMode);
    }

    if (this.lock) {
      throw new QueryBuilderUsageError('lock has already been set in that query');
    }

    this.lock = lockMode;
  };

  _proto2.buildSelect = function buildSelect() {
    // selecting no columns will result in empts-string selects which must be filtered
    return 'SELECT ' + this.select.filter(function (s) {
      return !!s;
    }).join(',');
  };

  _proto2.buildFrom = function buildFrom() {
    if (!this.from.length) {
      assert.fail('from must not be empty');
    }

    return 'FROM ' + this.from.join(',');
  };

  _proto2.buildJoin = function buildJoin() {
    return this.joins.map(function (j) {
      var joinType = j.joinType === 'join' ? 'JOIN' : 'LEFT JOIN';
      return joinType + " " + j.tableSql2 + " ON " + j.columnSql1 + " = " + j.columnSql2;
    }).join(' ');
  };

  _proto2.buildWhere = function buildWhere() {
    if (!this.where.length) {
      return '';
    }

    return 'WHERE ' + this.where.join(' AND ');
  };

  _proto2.buildGroupBy = function buildGroupBy() {
    if (!this.groupBy.length) {
      return '';
    }

    return 'GROUP BY ' + this.groupBy.join(',');
  };

  _proto2.buildOrderBy = function buildOrderBy() {
    if (!this.orderBy.length) {
      return '';
    }

    return 'ORDER BY ' + this.orderBy.join(',');
  };

  _proto2.buildLimit = function buildLimit() {
    if (this.limit === undefined) {
      return '';
    }

    return "LIMIT " + this.limit;
  };

  _proto2.buildOffset = function buildOffset() {
    if (this.offset === undefined) {
      return '';
    }

    return "OFFSET " + this.offset;
  };

  _proto2.buildLock = function buildLock() {
    if (!this.lock) {
      return '';
    }

    var lockStatement;

    switch (this.lock) {
      case 'share':
        lockStatement = 'SHARE';
        break;

      case 'update':
        lockStatement = 'UPDATE';
        break;

      case 'none':
        // disable locking, especially when using lockParam
        return '';

      default:
        assertNever();
    }

    return 'FOR ' + lockStatement;
  };

  _proto2.build = function build() {
    return this.buildSelect() + '\n' + this.buildFrom() + '\n' + this.buildJoin() + '\n' + this.buildWhere() + '\n' + this.buildGroupBy() + '\n' + this.buildOrderBy() + '\n' + this.buildLimit() + '\n' + this.buildOffset() + '\n' + this.buildLock();
  };

  _proto2.buildUpdate = function buildUpdate(table, columnsToSet, // actually keyof TableImplementation['tableColumns']
  columnsCtx, returning) {
    if (this.from.length !== 1) {
      assert.fail('need exactly one from clause');
    } // first all clauses that affect parameters


    var where = this.buildWhere(); // then add the column-set-clauses and let them use a separate set of parameters

    columnsCtx.setParameterOffset(this.ctx.getParameterMapping().length);
    var update = columnsToSet.map(function (c) {
      var param = columnsCtx.getNextParameter(c);
      var name = table[c].name;
      return name + '=' + param;
    }).join(',');
    return 'UPDATE ' + this.from[0] + ' SET ' + update + ' ' + where + ' RETURNING ' + returning;
  };

  return SqlQuery;
}(); // returns a function which will convert a row from the database into what is
// actually declared in the schema, e.g. string-formatted-dates (bc they where
// selected via a json query) into real javascript Date objects.


function buildResultConverter(query) {
  var converters = [];
  query.forEach(function (item) {
    switch (item.queryType) {
      case 'from':
        {
          var table = item.table;
          converters.push(table.getResultConverter());
        }
        break;

      case 'join':
        {
          var table2 = item.column2; // table 1 is already present in the query so we don't need to add its
          // result converter again

          if (item.joinType === 'join') {
            converters.push(table2.getResultConverter());
          } else if (item.joinType === 'leftJoin') {
            converters.push(table2.getResultConverter()); // TODO: pass nullable???
          } else {
            assertNever();
          }

          break;
        }
      // ignore all other query item types because they do not introduce new
      // tables/columns into the query

      case 'limit':
      case 'lock':
      case 'lockParam':
      case 'offset':
      case 'orderBy':
      case 'whereEq':
      case 'whereIn':
      case 'whereSql':
      case 'canaryColumn':
        break;

      default:
        assertNever();
    }
  });
  return function (row) {
    return converters.forEach(function (c) {
      return c(row);
    });
  };
}
function buildSqlQuery(query, ctx, // required as it may contain a param to determine locking
// TODO: maybe in the future it will also contain parameters for query
// objects which are enabled/disabled depending on a parameter
params) {
  var sql = new SqlQuery(ctx);
  var groupByNeeded = false;
  var groupByTables = new Set();
  query.forEach(function (item) {
    switch (item.queryType) {
      case 'from':
        {
          var table = item.table;
          var alias = sql.getAlias(table.tableName);
          sql.addFrom(table.getTableSql(alias, ctx, params));
          sql.addSelect(table.getSelectSql(alias, false));
          groupByTables.add(table);
        }
        break;

      case 'join':
        {
          var table1 = item.column1,
              table2 = item.column2,
              joinType = item.joinType;
          var alias1 = sql.getAlias(table1.tableName);
          var alias2 = sql.getAlias(table2.tableName);
          sql.addJoin(item.joinType, table2.getTableSql(alias2, ctx, params), table1.getReferencedColumnSql(alias1), table2.getReferencedColumnSql(alias2));
          sql.addSelect(table2.getSelectSql(alias2, joinType === 'leftJoin'));

          if (table2.isJsonAggProjection()) {
            groupByNeeded = true;
          } else {
            groupByTables.add(table2);
          }

          break;
        }

      case 'whereEq':
        {
          var _table = item.column;

          var _alias = sql.getAlias(_table.tableName);

          sql.addWhereEq(_table.getReferencedColumnSql(_alias), item.paramKey, !!_table.getReferencedColumn().isNullable);
          break;
        }

      case 'whereIn':
        {
          var _table2 = item.column;

          var _alias2 = sql.getAlias(_table2.tableName);

          sql.addWhereIn(_table2.getReferencedColumnSql(_alias2), item.paramKey);
          break;
        }

      case 'whereSql':
        {
          // aliases for each table referenced in the fragment
          var columnsSql = item.fragments.map(function (f) {
            if (!f.column) {
              return;
            }

            var table = f.column;
            var alias = sql.getAlias(table.tableName);
            return table.getReferencedColumnSql(alias);
          });
          sql.addWhereSql(item.fragments, columnsSql);
          break;
        }

      case 'orderBy':
        {
          var _table3 = item.column;

          var _alias3 = sql.getAlias(_table3.tableName);

          sql.addOrderBy(_table3.getReferencedColumnSql(_alias3), item.direction, item.nulls);
          break;
        }

      case 'limit':
        sql.setLimit(item.count);
        break;

      case 'offset':
        sql.setOffset(item.offset);
        break;

      case 'lock':
        sql.setLock(item.lockMode);
        break;

      case 'lockParam':
        {
          if (params) {
            var lockMode = params[item.paramKey];
            sql.setLock(lockMode);
            break;
          } else {
            throw new QueryBuilderUsageError('lockParam: parameter to determine lock mode is missing or empty');
          }
        }

      case 'canaryColumn':
        sql.addSelect("true AS " + item.columnName);
        break;

      default:
        assertNever();
    }
  }); // group by for json aggregations

  if (groupByNeeded) {
    groupByTables.forEach(function (table) {
      var alias = sql.getAlias(table.tableName);
      sql.addGroupBy(table.getPrimaryColumnsSql(alias));
    });
  }

  return sql.build();
} // return the columns to select when building subselects

function buildColumns(query) {
  var columns = {};
  query.forEach(function (item) {
    switch (item.queryType) {
      case 'from':
        {
          var table = item.table;
          Object.assign(columns, table.getColumns());
          break;
        }

      case 'join':
        {
          var table2 = item.column2; // TODO: PASS INFO WETHER ITS A LEFT  JOIN (NULLABLE)

          Object.assign(columns, table2.getColumns());
          break;
        }
      // ignore all other query item types, we only care about the ones that
      // modify column type information

      case 'limit':
      case 'lock':
      case 'lockParam':
      case 'offset':
      case 'orderBy':
      case 'whereEq':
      case 'whereIn':
      case 'whereSql':
      case 'canaryColumn':
        break;

      default:
        assertNever();
    }
  });
  return columns;
} // The insert statement must be tailored to the data we want to insert because
// node-postgres does not allow to pass a default value for missing
// (==undefined) column values (its using null instead).

function buildInsert(table, data) {
  // collect all present columns
  var columnSet = {};
  data.forEach(function (row) {
    for (var k in row) {
      if (row.hasOwnProperty(k)) {
        columnSet[k] = true;
      }
    }
  });
  var columns = Object.keys(columnSet); // Build the parameter placeholder value lists.
  // The value list must contain the values for each inserted column in the
  // *same* order. Also, according to our types, some rows may omit columns
  // with default values - insert an SQL-`DEFAULT` in this case.

  var insertParams = []; // placeholders: $n or DEFAULT

  var insertValues = []; // the actual values

  var paramCount = 0;
  data.forEach(function (row) {
    var rowParams = [];
    columns.forEach(function (col) {
      if (row.hasOwnProperty(col) && row[col] !== undefined) {
        paramCount += 1;
        rowParams.push('$' + paramCount);
        insertValues.push(row[col]);
      } else {
        // assume that undefined always means 'use the default'
        rowParams.push('DEFAULT');
      }
    });
    insertParams.push(rowParams.join(','));
  });
  var insertStatement = 'INSERT INTO "' + table.tableName + '" (' + columns.map(function (k) {
    return '"' + table.tableColumns[k].name + '"';
  }).join(',') + ') VALUES (' + insertParams.join('),(') + ') RETURNING ' + table.getSelectSql(undefined, false);
  return [insertStatement, insertValues];
}
function buildUpdate(query, paramsCtx, columnsToSet, dataCtx) {
  var sql = new SqlQuery(paramsCtx);
  var table;
  query.forEach(function (item) {
    switch (item.queryType) {
      case 'from':
        {
          table = item.table;

          var _alias4 = sql.getAlias(item.table.tableName);

          sql.addFrom(item.table.getTableSql(_alias4, paramsCtx));
        }
        break;

      case 'whereEq':
        {
          var _table4 = item.column;

          var _alias5 = sql.getAlias(_table4.tableName);

          sql.addWhereEq(_table4.getReferencedColumnSql(_alias5), item.paramKey, !!_table4.getReferencedColumn().isNullable);
        }
        break;

      case 'whereIn':
        {
          var _table5 = item.column;

          var _alias6 = sql.getAlias(_table5.tableName);

          sql.addWhereIn(_table5.getReferencedColumnSql(_alias6), item.paramKey);
          break;
        }

      case 'whereSql':
        {
          // aliases for each table referenced in the fragment
          var columnsSql = item.fragments.map(function (f) {
            if (!f.column) {
              return;
            }

            var table = f.column;
            var alias = sql.getAlias(table.tableName);
            return table.getReferencedColumnSql(alias);
          });
          sql.addWhereSql(item.fragments, columnsSql);
          break;
        }

      case 'join':
      case 'orderBy':
      case 'limit':
      case 'offset':
      case 'lock':
      case 'lockParam':
      case 'canaryColumn':
        throw new QueryBuilderUsageError("queryType is not allowed in updates: " + item.queryType);

      default:
        assertNever();
    }
  });

  if (!table) {
    assert.fail('table is missing in update');
  }

  var alias = sql.getAlias(table.tableName);
  var returning = table.getSelectSql(alias, false);
  return sql.buildUpdate(table.tableColumns, columnsToSet, dataCtx, returning);
}

/**
 * Tracks the dollar parameters and maps them to keys.
 *
 * Postgres only supports positional arguments in queries. So we need to
 * track them when the query is built at the end as we want our parameter type
 * to be an object so it works across subqueries and is mergable.
 */
var BuildContext =
/*#__PURE__*/
function () {
  function BuildContext() {
    this.offset = 0;
    this.mutableParameterMapping = [];
  }

  var _proto = BuildContext.prototype;

  _proto.setParameterOffset = function setParameterOffset(offset) {
    this.offset = offset;
  };

  _proto.getNextParameter = function getNextParameter(name) {
    this.mutableParameterMapping.push(name);
    return '$' + (this.mutableParameterMapping.length + this.offset).toString();
  };

  _proto.getParameterMapping = function getParameterMapping() {
    return this.mutableParameterMapping;
  };

  _proto.getParameters = function getParameters(paramObject) {
    return this.mutableParameterMapping.map(function (k) {
      return paramObject[k];
    });
  };

  return BuildContext;
}();

var sqlImplementation = function sqlImplementation(literals, param1, param2) {
  var column;
  var paramKey;
  var columnFirst = false;

  if (param1 && param1.__typesafQueryBuilderSqlFragmentParam === true) {
    paramKey = param1.paramKey;
    columnFirst = false;

    if (param2 instanceof TableImplementation) {
      column = param2;
    } else {
      if (param2 !== undefined) {
        assert__default.fail('expected param2 to be undefined');
      }
    }
  } else if (param1 instanceof TableImplementation) {
    column = param1;
    columnFirst = true;

    if (param2 && param2.__typesafQueryBuilderSqlFragmentParam === true) {
      paramKey = param2.paramKey;
    } else {
      if (param2 !== undefined) {
        assert__default.fail('expected param2 to be undefined');
      }
    }
  } else if (param1 === undefined && param2 === undefined) ; else {
    assert__default.fail("no matching parameters in sql fragment: " + literals);
  }

  return {
    column: column,
    columnFirst: columnFirst,
    paramKey: paramKey,
    // the paramValue attribute is only used in the typesystem
    literals: literals.raw
  };
};

function createSqlParam(key) {
  return {
    __typesafQueryBuilderSqlFragmentParam: true,
    paramKey: key
  };
}

sqlImplementation.param = createSqlParam;
sqlImplementation.number = createSqlParam;
sqlImplementation.string = createSqlParam;
sqlImplementation["boolean"] = createSqlParam;
sqlImplementation.date = createSqlParam;
sqlImplementation.numberArray = createSqlParam;
sqlImplementation.stringArray = createSqlParam;
var sql = sqlImplementation; // call each columns validation function for the given data and assign the


var uniqueTableNameCounter = 0;

var QueryImplementation =
/*#__PURE__*/
function () {
  function QueryImplementation(tables, query) {
    this.tables = tables;
    this.query = query;
    this.tables = tables;
    this.query = query; // TODO: raise an error if table1 has selected cols that differ from the selection that is used in the table
    // reason: catch usage errors where one tries pass a tablecolref with selected columns to join, because they get ignored
    // but it must be allowed to store a table with selections in a variable and then use it also in the join again

    this.checkSingleJsonAgg();
    this.checkDuplicateSelectedColumns();
  }

  var _proto = QueryImplementation.prototype;

  _proto.checkSingleJsonAgg = function checkSingleJsonAgg() {
    // raise an error if selectAsJsonAgg is used more than once in a query
    if (this.tables.length > 1 && this.tables.filter(function (t) {
      return t.isJsonAggProjection();
    }).length > 1) {
      throw new QueryBuilderUsageError('`selectAsJsonAgg` must only be used once in each query (use subqueries in case you want to have multiple `selectAsJsonAgg` aggregations)');
    }
  } // raise an error if selected columns of different conlfict with each other
  // (e.g. multiple id cols)
  ;

  _proto.checkDuplicateSelectedColumns = function checkDuplicateSelectedColumns() {
    if (this.tables.length < 2) {
      return;
    }

    var columnMultiset = new Map();
    this.tables.forEach(function (t) {
      t.getResultingColumnNames().forEach(function (c) {
        var entry = columnMultiset.get(c);

        if (entry) {
          entry.add(t);
        } else {
          columnMultiset.set(c, new Set([t]));
        }
      });
    });
    var containsDuplicate = Array.from(columnMultiset.values()).some(function (s) {
      return s.size > 1;
    });

    if (!containsDuplicate) {
      return;
    } // table-names -> list of duplicated column names


    var duplicatesReport = new Map();
    columnMultiset.forEach(function (tableSet, columnName) {
      if (tableSet.size < 2) {
        // column is just present in a single table
        return;
      }

      var reportKeyList = [];
      tableSet.forEach(function (t) {
        reportKeyList.push(t.tableName);
      });
      var reportKey = reportKeyList.join(', ');
      var entry = duplicatesReport.get(reportKey);

      if (!entry) {
        duplicatesReport.set(reportKey, [columnName]);
      } else {
        entry.push(columnName);
      }
    });
    var msg = [];
    duplicatesReport.forEach(function (cols, tables) {
      msg.push("in tables " + tables + ": " + cols.join(', '));
    });
    throw new QueryBuilderResultError("Ambiguous selected column names " + msg.join(' and '));
  };

  _proto.join = function join(ref1, ref2) {
    var table1 = getTableImplementation(ref1);
    var table2 = getTableImplementation(ref2);
    return new QueryImplementation([].concat(this.tables, [table2]), [].concat(this.query, [{
      queryType: 'join',
      column1: table1,
      column2: table2,
      joinType: 'join'
    }]));
  };

  _proto.leftJoin = function leftJoin(ref1, ref2) {
    var table1 = getTableImplementation(ref1);
    var table2 = getTableImplementation(ref2);
    return new QueryImplementation([].concat(this.tables, [table2]), [].concat(this.query, [{
      queryType: 'join',
      column1: table1,
      column2: table2,
      joinType: 'leftJoin'
    }]));
  };

  _proto.whereEq = function whereEq(column, paramKey) {
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'whereEq',
      column: getTableImplementation(column),
      paramKey: paramKey
    }]));
  };

  _proto.whereIn = function whereIn(column, paramKey) {
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'whereIn',
      column: getTableImplementation(column),
      paramKey: paramKey
    }]));
  };

  _proto.whereSql = function whereSql() {
    for (var _len = arguments.length, params = new Array(_len), _key = 0; _key < _len; _key++) {
      params[_key] = arguments[_key];
    }

    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'whereSql',
      fragments: params.map(function (f) {
        return {
          column: f.column ? getTableImplementation(f.column) : undefined,
          columnFirst: f.columnFirst,
          literals: f.literals,
          paramKey: f.paramKey
        };
      })
    }]));
  };

  _proto.limit = function limit(count) {
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'limit',
      count: count
    }]));
  };

  _proto.offset = function offset(_offset) {
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'offset',
      offset: _offset
    }]));
  };

  _proto.orderBy = function orderBy(column, direction, nulls) {
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'orderBy',
      column: getTableImplementation(column),
      direction: direction,
      nulls: nulls
    }]));
  };

  _proto.lock = function lock(lockMode) {
    // Does not work with json-aggregate columns at the moment bc they introduce a group-by.
    // TODO: In this case, we need to add the FOR UPDATE in a plain subselect of the json-aggregated table.
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'lock',
      lockMode: lockMode
    }]));
  };

  _proto.lockParam = function lockParam(paramKey) {
    return new QueryImplementation(this.tables, [].concat(this.query, [{
      queryType: 'lockParam',
      paramKey: paramKey
    }]));
  };

  _proto.table = function table() {
    var _this = this;

    uniqueTableNameCounter += 1; // table name which does not clash with real tables

    var tableName = '__typesafe_query_builder_' + uniqueTableNameCounter;
    var tableImplementation = new TableImplementation(tableName, buildColumns(this.query)); // to be able to generate postgres positional arguments and map them to
    // the `params: P` object we need delay building the sql until we know all
    // parameters

    tableImplementation.tableQuery = function (ctx, params, canaryColumnName) {
      var query = !canaryColumnName ? _this.query : [{
        queryType: 'canaryColumn',
        columnName: canaryColumnName
      }].concat(_this.query);
      return buildSqlQuery(query, ctx, params);
    };

    return tableImplementation.getTableProxy();
  };

  _proto.buildSql = function buildSql(ctx, params) {
    return buildSqlQuery(this.query, ctx || new BuildContext(), params);
  };

  _proto.sql = function sql(params) {
    return this.buildSql(undefined, params);
  };

  _proto.explain = function explain(client, params) {
    try {
      var _this3 = this;

      var ctx = new BuildContext();

      var _sql = 'EXPLAIN ' + _this3.buildSql(ctx, params);

      var paramArray = params ? ctx.getParameters(params) : [];
      return Promise.resolve(client.query(_sql, paramArray)).then(function (_client$query) {
        return _client$query.rows.map(function (r) {
          return r['QUERY PLAN'];
        }).join('\n');
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.explainAnalyze = function explainAnalyze(client, params) {
    try {
      var _this5 = this;

      var ctx = new BuildContext();

      var _sql2 = 'EXPLAIN ANALYZE ' + _this5.buildSql(ctx, params);

      var paramArray = params ? ctx.getParameters(params) : [];
      return Promise.resolve(client.query(_sql2, paramArray)).then(function (_client$query2) {
        return _client$query2.rows.map(function (r) {
          return r['QUERY PLAN'];
        }).join('\n');
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetch = function fetch(client, params) {
    try {
      var _this7 = this;

      var ctx = new BuildContext();

      var _sql3 = _this7.buildSql(ctx, params);

      var paramArray = params ? ctx.getParameters(params) : [];
      var resultConverter = buildResultConverter(_this7.query);
      return Promise.resolve(client.query(_sql3, paramArray)).then(function (_client$query3) {
        var result = _client$query3.rows;
        result.forEach(function (row) {
          return resultConverter(row);
        });
        return result;
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetchOne = function fetchOne(client, params) {
    try {
      var _this9 = this;

      var ctx = new BuildContext();

      var _sql4 = _this9.buildSql(ctx, params);

      var paramArray = params ? ctx.getParameters(params) : [];
      var resultConverter = buildResultConverter(_this9.query);
      return Promise.resolve(client.query(_sql4, paramArray)).then(function (_client$query4) {
        var rows = _client$query4.rows;

        if (!rows.length) {
          return;
        }

        if (rows.length > 1) {
          throw new QueryBuilderResultError("expected at most one row but the query returned: " + rows.length);
        }

        resultConverter(rows[0]);
        return rows[0];
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.fetchExactlyOne = function fetchExactlyOne(client, params) {
    try {
      var _this11 = this;

      var ctx = new BuildContext();

      var _sql5 = _this11.buildSql(ctx, params);

      var paramArray = params ? ctx.getParameters(params) : [];
      var resultConverter = buildResultConverter(_this11.query);
      return Promise.resolve(client.query(_sql5, paramArray)).then(function (_client$query5) {
        var rows = _client$query5.rows;

        if (rows.length !== 1) {
          throw new QueryBuilderResultError("expected exactly one row but the query returned: " + rows.length);
        }

        resultConverter(rows[0]);
        return rows[0];
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.use = function use(factory) {
    return factory(this);
  } // DML methods
  ;

  _proto.validateRowsData = function validateRowsData(table, data) {
    // keep track of the current validation location so we can use one big
    // try-catch and don't have to wrap every column validator which might be
    // inefficient for large inserts
    var currentRowIndex = 0;
    var currentRow = undefined;
    var currentKey = '';

    try {
      // validate the data of each column before insertion
      data.forEach(function (row, i) {
        currentRow = row;
        currentRowIndex = i;
        var keys = Object.keys(row);
        keys.forEach(function (k) {
          currentKey = k;
          var value = row[k];
          var column = table.tableColumns[k];

          if (!column) {
            assert__default.fail('column is missing from table implementation ' + k);
          } // throws on invalid data


          row[k] = column.columnValue(value);
        });
      });
    } catch (e) {
      // Provide additional context info when the validation fails.
      // Otherwise its next to impossible to find the invalid column esp for
      // large tables with many custom (json) runtypes.
      var msg = "validation failed for column " + JSON.stringify(currentKey) + " at row number " + currentRowIndex + " with: " + JSON.stringify(e.message);
      throw new QueryBuilderValidationError(msg, table.tableName, currentKey, currentRowIndex, currentRow, e);
    }
  };

  _proto.insert = function insert(client, data) {
    try {
      var _this13 = this;

      if (!data.length) {
        return Promise.resolve([]);
      }

      if (_this13.tables.length !== 1) {
        // this is actually prohibited by the type system
        assert__default.fail('expected exactly one table');
      } // validate and sanitize data in place according to the tables column
      // definitions


      _this13.validateRowsData(_this13.tables[0], data);

      var _buildInsert = buildInsert(_this13.tables[0], data),
          _sql6 = _buildInsert[0],
          insertValues = _buildInsert[1];

      return Promise.resolve(client.query(_sql6, insertValues)).then(function (_client$query6) {
        return _client$query6.rows;
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.insertOne = function insertOne(client, data) {
    try {
      var _this15 = this;

      return Promise.resolve(_this15.insert(client, [data])).then(function (_this14$insert) {
        return _this14$insert[0];
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.update = function update(client, params, data) {
    try {
      var _this17 = this;

      if (_this17.tables.length !== 1) {
        // this is actually prohibited by the type system
        assert__default.fail('expected exactly one table');
      }

      var table = _this17.tables[0];
      var paramsCtx = new BuildContext(); // parameters for the `WHERE` conditions

      var dataCtx = new BuildContext(); // the actual values for the update

      var dataColumns = Object.keys(data);

      if (!dataColumns.length) {
        // nothing to update
        return Promise.resolve([]);
      } // validate and sanitize data in place according to the tables column
      // definitions


      _this17.validateRowsData(table, [data]);

      var columns = Object.keys(data);

      var _sql7 = buildUpdate(_this17.query, paramsCtx, columns, dataCtx);

      return Promise.resolve(client.query(_sql7, paramsCtx.getParameters(params).concat(dataCtx.getParameters(data)))).then(function (_client$query7) {
        return _client$query7.rows;
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.updateOne = function updateOne(client, params, data) {
    try {
      var _this19 = this;

      return Promise.resolve(_this19.update(client, params, data)).then(function (rows) {
        if (rows.length > 1) {
          throw new QueryBuilderResultError("expected at most one updated row but the query updated: " + rows.length);
        }

        return rows;
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  _proto.updateExactlyOne = function updateExactlyOne(client, params, data) {
    try {
      var _this21 = this;

      return Promise.resolve(_this21.update(client, params, data)).then(function (rows) {
        if (rows.length !== 1) {
          throw new QueryBuilderResultError("expected exactly one updated row but the query updated: " + rows.length);
        }

        return rows;
      });
    } catch (e) {
      return Promise.reject(e);
    }
  };

  return QueryImplementation;
}();
/**
 * Chaining API root.
 */


function query(table) {
  var ti = getTableImplementation(table);
  return new QueryImplementation([ti], [{
    queryType: 'from',
    table: ti
  }]);
}

/**
 * Pick defined keys from an object returning a new object.
 */
function pick(obj) {
  var res = {};

  for (var _len = arguments.length, keys = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    keys[_key - 1] = arguments[_key];
  }

  for (var i = 0; i < keys.length; i++) {
    if (obj.hasOwnProperty(keys[i])) {
      // Only pick keys that are present on the source object.
      // Same behavior as lodash.pick.
      // Allows picking keys from Partial objects without introducing
      // undefined values in the resulting object
      // e.g.:
      //   pick({}, 'a') is {} and not {a: undefined}
      // and:
      //   pick({a: undefined}, 'a') is {a: undefined} and not {}
      res[keys[i]] = obj[keys[i]];
    }
  }

  return res;
}
/**
 * Omit keys from an object returning a new object.
 */

function omit(obj) {
  var res = _extends({}, obj);

  for (var _len2 = arguments.length, keys = new Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
    keys[_key2 - 1] = arguments[_key2];
  }

  for (var i = 0; i < keys.length; i++) {
    delete res[keys[i]];
  }

  return res;
}

exports.Column = Column;
exports.QueryBuilderError = QueryBuilderError;
exports.QueryBuilderResultError = QueryBuilderResultError;
exports.QueryBuilderUsageError = QueryBuilderUsageError;
exports.QueryBuilderValidationError = QueryBuilderValidationError;
exports.column = column;
exports.omit = omit;
exports.pick = pick;
exports.query = query;
exports.sql = sql;
exports.table = table;
//# sourceMappingURL=typesafe-query-builder.cjs.development.js.map
